// Utena IIIMP Server for Anthy
//
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/poll.h>
#include <sys/wait.h>
#include <netinet/in.h>
#include <netdb.h>
#include <signal.h>
#include <unistd.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <iconv.h>
#include <list>
#include <anthy.h>
#include "rkconv.h"
#include "rkmap.h"
#include "imconn.h"
#include "packet.h"

#define PORT_NUMBER 9010

int iiim_trace_flag;
int trace_flag;
static int background_p;

RKMap *rk_hira_map;
RKMap *rk_kata_map;

static int port_number=PORT_NUMBER;
static list<IMConn *> ims;

extern void initKeyAtoms();

static void parse_args(int argc,char **argv);
static void print_usage();
static void print_version();
static int event_loop();
static int accept_proc(int fd);
static void kill_prev();
static void go_background();


#if __GLIBC_MINOR__ == 1 || FreeBSD
#define ICONV_ARG2 (const char **)
#else
// conforming to new standard draft 5/2000
#define ICONV_ARG2
#endif

static char *sa2str (struct sockaddr*, socklen_t); 

static list<int> serv_fds;

static int init_serv_fd(char *hostname,char *port);

STRING createSTRING(char *s)
{
  STRING str;
  iconv_t conv;

  conv = iconv_open("UTF-16","EUC-JP");
  if ( (int)conv == -1 ){
    printf("Failed to create iconv.\n");
    exit(0);
  }

  size_t ilen,olen,buflen,ret,len;
  int i;
  len = strlen(s);
  ilen = len;
  buflen = ilen*5;
  char *outbuf = (char *)alloca(buflen);
  char *pp = outbuf;
  ret = iconv(conv,ICONV_ARG2 &s,&ilen,&outbuf,&buflen);
  if ( ret == -1 ){
    printf("failed iconv.\n");
    return str;
  }
  unsigned short *p;
  p = (unsigned short*)pp;
  olen = len*5 - buflen;

  for ( i = 0 ; i < olen /2 ; i++){
    // Is UTF BIG ENDIAN?
    str.push_back((p[i]>>8)|((p[i]<<8)&65280));
  }
  iconv_close(conv);
  return str;
}

void kill_prev()
{
  char *s;
  s = getenv("UTENA_PID");
  if ( s ){
    pid_t p = atoi(s);
    if ( p ){
      kill(p,SIGKILL);
    }
  }
  printf("Can't find Utena. (in this world)\n");
}

int init_serv_fd(char *hostname,char *port)
{
  int foo;
  int s;
#ifdef INET6
  struct in_addr ia;
  struct addrinfo hints, *res, *ai; 
 
  memset(&hints, 0, sizeof(hints)); 
  hints.ai_family = AF_UNSPEC; 
  hints.ai_socktype = SOCK_STREAM; 
  hints.ai_protocol = IPPROTO_TCP; 
  hints.ai_flags = AI_PASSIVE; 
 
  if (getaddrinfo(NULL, port, &hints, &res) != 0) {
    return -1; 
  }
  
  for (ai = res; ai != NULL; ai= ai->ai_next) { 
    if ((s = socket(ai->ai_family, ai->ai_socktype, 
		    ai->ai_protocol)) < 0) {
      continue; 
    }
 
    foo = 1;
    foo = setsockopt(s,SOL_SOCKET,SO_REUSEADDR,&foo ,4);
    if ( foo == -1){
      // IGNORE. No problem.
      //perror("failed to set reuse option");
    }

    if ( bind(s, ai->ai_addr, ai->ai_addrlen) < 0) {
      close(s); 
      continue; 
    } 
 
    if (listen(s,1) < 0) { 
      close(s); 
      continue; 
    } 
 
    serv_fds.push_back(s); 
  } 

  if (res != NULL) {
    freeaddrinfo(res);
  }

#else /*!INET6*/

  struct sockaddr_in myhost;

  s = socket(AF_INET,SOCK_STREAM,0);
  if (  s == -1 ){
    perror("failed in socket()");
    return -1;
  }

  foo = 1;
  foo = setsockopt(s,SOL_SOCKET,SO_REUSEADDR,&foo ,4);
  if ( foo == -1){
    // IGNORE. No problem.
    //perror("failed to set reuse option");
    //return -1;
  }

  myhost.sin_family=AF_INET;
  myhost.sin_port=htons(port_number);
  myhost.sin_addr.s_addr=htonl(INADDR_ANY);

  foo = bind(s,(struct sockaddr *)&myhost,16);
  if( foo == -1 ){
    perror("failed in bind()");
    return -1;
  }
  foo = listen(s,5);
  if ( foo == -1){
    perror("failed in listen()");
    return -1;
  } 
  serv_fds.push_back(s);
#endif
  if ( !serv_fds.size() ){
    return -1;
  }
  return 0;
}

void print_version()
{
  printf("UTENA(Anthy-server)-"VERSION"\n");
  exit(0);
}

void print_usage()
{
  printf("UTENA(Anthy-server)-"VERSION".\n"
	 " --trace: trace connection.\n"
	 " --trace-iiim: trace IIIM packets.\n"
	 " --version: show version.\n"
	 " --help: show this help.\n"
	 " --background: run background.\n"
	 " --kill: kill previous process.\n");
  exit(0);
}

void parse_args(int argc,char **argv)
{
  int i;
  for ( i = 1 ; i < argc ; i++){
    if ( !strncmp(argv[i],"--",2)){
      char *arg = &argv[i][2];
      if ( !strcmp(arg,"trace")){
	trace_flag = 1;
      }else if ( !strcmp(arg,"trace-iiim")){
	iiim_trace_flag = 1;
      }else if (!strcmp(arg,"help")){
	print_usage();
      }else if (!strcmp(arg,"version")){
	print_version();
      }else if (!strcmp(arg,"background")){
	background_p = 1;
      }else if (!strcmp(arg,"kill")){
	kill_prev();
	exit(0);
      }else{
	print_usage();
      }
    }
  }
}

int accept_proc(int fd)
{
  socklen_t len;
#ifdef INET6
  struct sockaddr_storage peer;
#else
  struct sockaddr peer;
#endif
  len = sizeof(peer);
  int new_fd = accept(fd,(struct sockaddr *)&peer,&len);
  if ( new_fd == -1 ){
    perror("Failed to accept.\n");
    return -1;
  }else{
    printf("Accepted TCP connection from %s\n",
	   sa2str((struct sockaddr *)&peer,len));
  }
  if ( trace_flag ){
    printf("Accept\n");
  }
  IMConn *conn = new IMConn(new_fd);
  ims.push_back(conn);
  return 0;
}

int event_loop()
{
  /* setup poll table */
  struct pollfd *polltab;
  int nr_fd ,nth,nr_accept_fd;
  nr_fd = serv_fds.size() + ims.size();
  polltab = (struct pollfd *)alloca(sizeof(struct pollfd)*nr_fd);
  nth = 0;

  /* setup polltab for fd to accept */
  {
    list<int>::iterator it;
    for ( it = serv_fds.begin() ; it != serv_fds.end() ; it ++ ){
      polltab[nth].fd = *it;
      polltab[nth].events = POLLIN;
      nth++;
    }
  }

  /* setup polltab for each connection */
  {
    list<IMConn *>::iterator it;
    for ( it = ims.begin() ; it != ims.end() ; it ++){
      polltab[nth].fd = (*it)->getFd();
      polltab[nth].events = POLLIN;
      if ( (*it)->isWritePending()){
	polltab[nth].events |= POLLOUT;
      }
      nth++;
    }
  }

  /* do poll */
  poll(polltab,nth,1000);

  /* check result for each connection */
  nth = 0;
  /*accept fds*/
  {
    list<int>::iterator it;
    for ( it = serv_fds.begin() ; it != serv_fds.end() ; it ++){
      if ( polltab[nth].revents & POLLIN ){
	accept_proc(*it);
      }
      nth++;
    }
  }
  {
    /*established connections*/
    list<IMConn *>::iterator it;
    for ( it = ims.begin() ; it != ims.end() && nth< nr_fd ; it ++){
      if ( polltab[nth].revents & POLLIN ){
	(*it)->read_proc();
      }
      if ( polltab[nth].revents & POLLOUT ){
	(*it)->write_proc();
      }
      nth ++;
    }
  }

  /* reap invalid connection */
  {
    list<IMConn *>::iterator it;
    for ( it = ims.begin() ; it != ims.end() ; ){
      if ( (*it)->isValid() ){
	it++;
      }else{
	if ( trace_flag ){
	  printf("Disconnect\n");
	}
	IMConn *im = *it;
	it = ims.erase(it);
	delete im;
      }
    }
  }
  return 0;
}

void go_background()
{
  pid_t pid = fork();
  if ( pid == -1 ){
    perror("Failed to do fork()");
    exit(-1);
  }else if(pid == 0 ){
    pid = fork();
    if ( pid ){
      exit(0);
    }else{
      /*worker process*/
      pid = getpid();
      printf("UTENA_PID=%d ;export UTENA_PID\n",pid);
      printf("echo Utena pid=%d\n",pid);
      fflush(stdout);
    }
  }else {
    /*original process*/
    wait(&pid);
    exit(0);
  }
}

int main(int argc,char **argv)
{
  int newfd,addr_len;
  struct sockaddr_in peer;

  parse_args(argc,argv);

  if ( background_p ){
    go_background();
  }

  initKeyAtoms();

  if ( anthy_init() == -1 ){
    printf("Failed to init anthy.\n");
    return -1;
  }
  rk_hira_map = new RKMap(hiragana);
  rk_kata_map = new RKMap(katakana);

  char portnum[5];
  char buf[NI_MAXSERV];
  sprintf(portnum,"%d",PORT_NUMBER);

  if (  init_serv_fd("localhost",portnum) < 0 ){
    printf("Failed to find usable address.\n"
	   " Maybe other instance is running.\n");
    exit(-1);
  }

  if ( trace_flag ){
    printf("Begin\n");
  }
  while(!event_loop());
  return 0;
}

static char *sa2str(struct sockaddr *sa, socklen_t salen) 
{ 
  static char retbuf[NI_MAXHOST + NI_MAXSERV + 3]; 
  char hbuf[NI_MAXHOST], sbuf[NI_MAXSERV]; 
 
#ifdef INET6 
  if(getnameinfo(sa, salen, hbuf, sizeof(hbuf), sbuf, sizeof(sbuf), 
                NI_NUMERICHOST | NI_NUMERICSERV) != 0) 
    return("???"); 
 
  sprintf(retbuf, "[%s]:%s", hbuf, sbuf); 
#else
  // IPv4 Only
  sprintf(retbuf,"[IPv4 Connection]");
#endif
  return retbuf;
}
